/**
	The Birds Update
	ASI Plugin	(Programming)	- by LINK/2012		(http://www.link2012world.blogspot.com/)
	New Birds	(Textures)		- by Junior_Djjr	(http://juniordjjr.blogspot.com.br/)
*/
//#define BIRDS_DEBUG
#define _CRT_SECURE_NO_WARNINGS

#include "NewBirds.h"
#include "CBirds.h"
#include "RwV3D.h"
#include "RwRGBA.h"
#include "RenderWare.h"
#include "CColLine.h"
#include "ColSphere.h"
#include "Addresses.h"
#include "NewBirds.h"
#include <cmath>
#include <cstdio>

#define BIRD_VERTEX_COUNT	9

#pragma warning(disable : 4482)

size_t CBirds::StatusSaver = 0;
CBird CBirds::BirdPool[];
CBird* CBirds::PoolEnd = nullptr;
RwV3D* CBirds::LastKilledBirdPos = nullptr;
bool* CBirds::HasKilledAnyBird = nullptr;
int* CBirds::nBirdsCreated = nullptr;
bool CBirds::IsInDebug = false;
float* CBirds::XOffset;
float* CBirds::YOffset;
float* CBirds::ZOffset;

static int BirdTXD = 0;

/*#########################################################################################*/

bool IsTimeInRange(int time, int a, int b)
{
	return
		a == b? (time == a):
		b > a? (time >= a && time < b):
			   (time >= a || time < b);
}

static void OnInitBirds()
{
	// Patch indices!
	unsigned int* indices = (unsigned int*)0x8D52F8;
	indices[27] = 8;
	indices[24] = 8;

	// Load birds.dat
	chdirProgramDir("");
	GetBirdsData("data\\birds.dat", BirdTypes);

	//
	*(void**)(0xC6AA54) = (void*)&CBirds::MethodTable;
}

static void OnUnInitBirds()
{
	DestroyBirdsDictionary(BirdTypes);
}

void* CBirds::MethodTable[] =
{
	// Original Methods Link
		(void*)(&CBirds::Init),				// [0]
		(void*)(&CBirds::UnInit),			// [1]
		(void*)(&CBirds::Add),				// [2]
		(void*)(&CBirds::Process),			// [3]
		(void*)(&CBirds::Render),			// [4]
		(void*)(&CBirds::CheckIfHitByFire),	// [5]

	// Variables Link
		(void*)(CBirds::BirdPool),			// [6]	- CBirds[]
		(void*)(&CBirds::PoolSize),			// [7]	- const int
		(void*)(&CBirds::IsInDebug),		// [8]	- bool*
		(void*)(sizeof(CBird)),				// [9]	- not a pointer
		(void*)(nullptr),					// [10]	- reserved
		(void*)(nullptr),					// [11]	- reserved
		(void*)(nullptr),					// [12]	- reserved

	// Additional Methods
		(void*)(&CBirds::DestroyBird),		// [13]
		(void*)(&CBirds::DestroyAllBirds),	// [14]
		(void*)(&CBirds::CreateDebugBird),	// [15]
		(void*)(&CBirds::RenderBirdNow),	// [16]
		(void*)(&CBirds::SetBirdtype),		// [17]
		(void*)(&CBirds::CreateDebugBird),	// [18]
		(void*)(&CBirds::AddNewBirdtype),	// [19]
		(void*)(&CBirds::GetBirdtypePtr),	// [20]
		(void*)(&CBirds::ReloadBirdTextures),// [21]
		(void*)(&CBirds::GetBirdIndexByIdentifier), // [22]
		(void*)(&CBirds::RewriteBirdData), // [23]
		(void*)(&CBirds::SaveDataStatus), // [24]
		(void*)(&CBirds::RestoreDataStatus), // [25]
		(void*)(&CBirds::ReloadBird),		 // [26]
		(void*)(&CBirds::SetBirdSize),		 // [27]
		(void*)(&CBirds::SetWingSpeed),		 // [28]
		(void*)(&CBirds::SetSpeed),			// [29]

	// ...
		(void*)(nullptr)					// [??]	
};


extern "C" __declspec(dllexport) void* GetCBirdsMethodTable()
{
	return CBirds::MethodTable;
}

/*##########################################################################################*/






void __cdecl CBirds::Init()
{
	for(CBird* bird = BirdPool; bird != PoolEnd; ++bird)
		bird->HasBirdHere = false;
	for(CBird* bird = BirdPool; bird != PoolEnd; ++bird)
		bird->HasBirdHere = false;

	*nBirdsCreated = 0;
	*HasKilledAnyBird = false;

	//--
	OnInitBirds();
}

void __cdecl CBirds::UnInit()
{
	for(CBird* bird = BirdPool; bird != PoolEnd; ++bird)
		bird->HasBirdHere = false;
	*nBirdsCreated = 0;

	//--
	OnUnInitBirds();
}


void __cdecl CBirds::CheckIfHitByFire(RwV3D* p1, RwV3D* p2)
{
	ColSphere BirdSphere;	// Bird Bounding Sphere
	CColLine Fire;			// Weapon Fire line

	CColLine__Constructor(&Fire, 0, p1, p2);

	for(CBird* bird = BirdPool; bird != PoolEnd; ++bird)
	{
		if(bird->HasBirdHere && !bird->HitByFire)
		{
			BirdSphere.sphere.center = bird->pos;
			BirdSphere.sphere.radius = 0.5;
			if(CColLine__HasCollisionWithSphere(&Fire, &BirdSphere))
			{
				*LastKilledBirdPos = bird->pos;
				CBirds::KillBird(bird);
				//*nBirdsCreated -= 1;
				//bird->HasBirdHere = false;
			}
		}
	}
}

void __cdecl CBirds::Process()
{
	RwV3D CamPos;

	// Get camera position
	if(camera->__parent.m_pCoords)
		CamPos = (reinterpret_cast<Placement*>(((char*)camera->__parent.m_pCoords + 0x30))->pos);
	else
		CamPos = camera->__parent.placement.pos;

	// Only create the bird if active interior is 0, has free slots in bird pool, hours is between 5~22,
	// and a bitwise operation with playingTime
	do
	{
		if(!CBirds::IsInDebug && *activeInterior == 0
			&& *nBirdsCreated < CBirds::PoolSize && (*playingTime & 0x1FF) == 6)
		{
#ifdef BIRDS_DEBUG
			PrintNow("Trying to spawn a random bird...", 1000, false);
#endif

			int birdType;
			int birdCount;
			float randHeight;
			float distance;
			long double angle;
			float waterLevel;
			CVector v;
			RwV3D BirdPos, BirdDest;

			// Find bird type to spawn
			{
				std::vector<size_t> Candidates;
				Candidates.reserve(BirdTypes.size());

				TownID CurrentTownID = (TownID)*wTownID;
				bool IsNearWater = camera->DistanceToWater < 30.0;

				// Find candidates
				for(size_t i = 0; i < BirdTypes.size(); ++i)
				{
					auto& Brd = BirdTypes[i];
					if(*Brd.Identifier == 0) continue;

					// Check basics conditions
					if(
						GetFlag(Brd.Flags, BIRD_SPAWN_NEAR_WATER) == IsNearWater
						&& Brd.SpawningTown[CurrentTownID] == true
						&& IsTimeInRange(*currentTimeHours, Brd.SpawnTime.Min, Brd.SpawnTime.Max)
					)
					{
						Candidates.push_back(i);
					}
				}

				// Raffle the candidates
				if(Candidates.size() == 0)
				{
					// No candidates, break

				#ifdef BIRDS_DEBUG
				{
					static char bufDebug[128]; 
					sprintf(bufDebug, "%d Failed to spawn bird, No candidates for this area...", BirdTypes.size());
					PrintNow(bufDebug, 2000, false);
				}
				#endif
					break;	
				}
				birdType = Candidates[random(0, Candidates.size())];
			}

			auto& Brd = BirdTypes[birdType];

			// Get number of birds in group
			birdCount = random(1, (CBirds::PoolSize + 1) - *nBirdsCreated);
			if(birdCount < Brd.BirdsInFlock.Min) birdCount = Brd.BirdsInFlock.Min;
			else if(birdCount >= Brd.BirdsInFlock.Max) birdCount = Brd.BirdsInFlock.Max - 1;
		
			if(birdCount > CBirds::PoolSize - *nBirdsCreated)
			{
				// Can't create the flock
				#ifdef BIRDS_DEBUG
					PrintNow("Failed to spawn bird, Birds pool is full!", 1000, false);
				#endif
				break; 
			}

			rand(); rand();
			distance = Brd.Distance;

			// Get X, Y Coord and Angle
			{
				if(rand() & 1)
				{
					angle = (long double)(rand() & 0xFF) * 0.02453125081956387;
				}
				else
				{
					v.x = camera->field_974.matrix.top.x;
					v.y = camera->field_974.matrix.top.y;
					v.z = 0;
					if(camera->field_974.matrix.top.x == 0.0)
						v.x = (float)(9.99999977648258209228515625e-3);
					CVector__normalize(&v);
					angle = (long double)((rand() & 0xFF) - 128) * 0.00390625 + atan2(v.y, v.x);
				}

				BirdPos.x = float( sin(angle) * (distance - 1.0)/*FIX*/ + CamPos.x );
				BirdPos.y = float( cos(angle) * (distance - 1.0)/*FIX*/ + CamPos.y );
			}


			// Get bird Z Coord
			{
				randHeight = randomf(Brd.Height.Min, Brd.Height.Max);

				if(GetFlag(Brd.Flags, BIRD_HEIGHT_RELATIVE_TO_CAMERA))
					randHeight += CamPos.z;
				else if(GetFlag(Brd.Flags, BIRD_HEIGHT_RELATIVE_TO_GROUND))
					randHeight += FindGroundZForCoord(BirdPos.x, BirdPos.y);
				else
				{
					// relative to 0.0f - not recommended - check if coordinate is above ground
					float ground = FindGroundZForCoord(BirdPos.x, BirdPos.y) + 1;
					if(randHeight <= ground)
					{
						#ifdef BIRDS_DEBUG
							PrintNow("Failed to spawn bird, bird in bluehell", 1000, false);
						#endif
						break; // Crap, I'm a bird in the bluehell
					}
				}

				if(randHeight >= Brd.Distance) randHeight = Brd.Distance - 1.0f;

				BirdPos.z = randHeight;
			}


			// Check if the position is not in the water
			if(!CWaterLevel__GetWaterLevelNoWaves(BirdPos.x, BirdPos.y, BirdPos.z, &waterLevel, nullptr, nullptr)
			|| waterLevel + 4.0 < BirdPos.z)
			{
				// Get bird destination

				v.x = camera->field_974.matrix.top.x;
				v.y = camera->field_974.matrix.top.y;
				v.z = 0;
				CVector__normalize(&v);

				BirdDest.x = (float)(CamPos.x + (v.x * 8.0));
				BirdDest.y = (float)(CamPos.y + (v.y * 8.0));
				BirdDest.z = BirdPos.z;

				#ifdef BIRDS_DEBUG
				{
					static int countDebug = -1; ++countDebug;
					static char bufDebug[128];
					sprintf(bufDebug, "(%d) Creating bird type %d counter %d", countDebug, birdType, birdCount);
					PrintNow(bufDebug, 5000, false);
				}
				#endif

				// Add birds to pool!
				CBirds::Add(BirdPos, BirdDest, birdCount, birdType, true);
			}
			else
			{
				#ifdef BIRDS_DEBUG
					PrintNow("Failed to spawn bird, too near the water level", 1000, false);
				#endif
			}

		}

	} while(false);

	// Remove birds out of vision
	{
		CBird* bird = &CBirds::BirdPool[ (*playingTime % CBirds::PoolSize)  ];
		if(bird->HasBirdHere)
		{
			float vx = bird->pos.x - CamPos.x;
			float vy = bird->pos.y - CamPos.y;

			if(  (bird->MaxBirdDistance * bird->MaxBirdDistance) < ((vy * vy) + (vx * vx)) )
			{
				if(!bird->IsDebugBird)
				{
					#ifdef BIRDS_DEBUG
						static char debugBuf[128];
						sprintf(debugBuf, "Removing out of vision bird %d", bird->m_Type);
						PrintNow(debugBuf, 2000, false);
					#endif
					DestroyBird(bird);
				}
			}

		}
	}

	// Update birds position
	{
		for(CBird* bird = CBirds::BirdPool; bird != CBirds::PoolEnd; ++bird)
		{
			if(bird->HasBirdHere && (!bird->IsDebugBird || bird->IsMoveableBird) &&
				bird->field_38 && *currentTime > (DWORD)bird->field_28)
			{
				if(bird->MustDoCurves)	//
				{
					long double v19 = *TimeStep * (!bird->HitByFire? 0.002000000094994903 : 0.05);
					long double v20 = bird->Velocity.x;
					long double v21 = bird->Velocity.y;
					long double v22 = sin(v19);
					long double v23 = cos(v19);

					bird->Velocity.x = (float)(v23 * v20 + v22 * v21);
					bird->Velocity.y = (float)(v23 * v21 - v22 * v20);
				}

				long double v25 = *TimeStep * 0.01999999955296516;
				
				//v35.x = bird->pos.x;
				//v35.y = bird->pos.y;
				//v35.z = bird->pos.z;

				long double v24 = 0.5 * v25;
				if(v24 > 1.0f) v24 = 1.0f;

				bird->field_C.x = (float)( (1.0 - v24) * bird->field_C.x + bird->Velocity.x * v24 );
				bird->field_C.y = (float)( (1.0 - v24) * bird->field_C.y + bird->Velocity.y * v24 );
				bird->field_C.z = (float)( (1.0 - v24) * bird->field_C.z + bird->Velocity.z * v24 );

				bird->pos.x += (float)(v25 * bird->field_C.x);
				bird->pos.y += (float)(v25 * bird->field_C.y);
				bird->pos.z += (float)(v25 * bird->field_C.z);

				bird->angle = atan2(bird->Velocity.x, bird->Velocity.y); // <--- This is right! Don't cry.

			}

		}
	}


}


int __cdecl CBirds::Add(RwV3D pos, RwV3D dest, int numBirds, int birdType, bool checkObstacles, ...)
{
	// Security checks
	{
		if(!numBirds) return 0;

		// New feature of CBirds::Add:
		// If birdType is -1 then we should have the bird identifier as optional\variable arg
		if(birdType == -1)
		{
			// Get variable arg
			va_list va; va_start(va, checkObstacles);
			char* sz = va_arg(va, char*);
			birdType = CBirds::GetBirdIndexByIdentifier(sz);
			va_end(va);

			// No bird found with identifier in last param
			if(birdType == -1)
			{
				#ifdef BIRDS_DEBUG
					PrintNow("Failed to spawn bird, invalid bird name", 1000, false);
				#endif
				return 0;
			}
		}

		if((unsigned)birdType >= BirdTypes.size())
		{
			#ifdef BIRDS_DEBUG
					PrintNow("Failed to spawn bird, invalid bird type", 1000, false);
			#endif
			return 0;
		}
	}

	float MaxBirdDistance;
	int i = -1;

	MaxBirdDistance = BirdTypes[birdType].Distance;

	for(int count = 0; count < numBirds; ++count)
	{
		// Optimized: No need to reloop in the pool, continue from where I've stopped
		++i;
		while(&BirdPool[i] < PoolEnd)
		{
			if(!BirdPool[i].HasBirdHere) break;
			++i;
		}

		if(&BirdPool[i] >= PoolEnd) return count;	// No free slot on the pool
		
		CBird* bird = &BirdPool[i];
		CVector v;

		bird->pos = pos;
		
		v.x = dest.x - bird->pos.x;
		v.y = dest.y - bird->pos.y;
		v.z = dest.z - bird->pos.z;
		CVector__normalize(&v);

		dest.x = (float)(v.x * MaxBirdDistance * 2.400000095367432 + bird->pos.x);
		dest.y = (float)(v.y * MaxBirdDistance * 2.400000095367432 + bird->pos.y);
		dest.z = (float)(v.z * MaxBirdDistance * 2.400000095367432 + bird->pos.z);

		if(checkObstacles
		&& !IsLineOfSightClear(&bird->pos, &dest, true, false, false, false, false, false, false))
		{
			#ifdef BIRDS_DEBUG
				PrintNow("WARNING: 1+ not created, obstacle", 1000, false);
			#endif
			continue;
		}


		bird->m_Type = birdType;

		int randSpeed = rand() % 31;
		bird->m_randSpeed = randSpeed;

		bird->HasBirdHere = true;

		bird->angle = atan2(v.x, v.y);	// <--- Don't cry, that's okay
		
		bird->field_38 = true;
		bird->field_28 = 0;
		bird->MustDoCurves = false;

		bird->pos.x += XOffset[count];
		bird->pos.y += YOffset[count];
		bird->pos.z += ZOffset[count];

		bird->HitByFire = false;
		bird->IsMoveableBird = true;
		bird->IsDebugBird = false;

		// Setup bird type specific options
		{
			auto& Type = BirdTypes[birdType];

			long double vx = (long double)randSpeed * 0.01999999955296516 + Type.Speed;
			
			bird->Velocity.x = (float)(v.x * vx);
			bird->Velocity.y = (float)(v.y * vx);
			bird->Velocity.z = (float)(v.z * vx);
			bird->field_C = bird->Velocity;

			if( Type.WingSpeedDiv == -1 || (GetFlag(Type.Flags, BIRD_GLIDABLE) && rand() & 4) )
				bird->WingSpeed = 1000000;
			else
			{
				bird->WingSpeed = Type.WingSpeedDiv > (randSpeed*6)? Type.WingSpeedDiv - randSpeed*6: Type.WingSpeedDiv;
			}

			//bird->MustDoCurves = (GetFlag(Type.Flags, BIRD_DO_CURVES) && rand() & 8);
			bird->MustDoCurves = false;

			bird->Size = randomf(Type.Size.Min, Type.Size.Max);
			bird->MaxBirdDistance = Type.Distance;

			bird->PolyColor1 = bird->PolyColor2 = bird->PolyColor3 =
			bird->PolyColor4 = bird->PolyColor5 = bird->PolyColor6 = 0xFF;
		}

		*CBirds::nBirdsCreated += 1;
	}


	return numBirds;
}



void __cdecl CBirds::Render()
{
	if(!*CBirds::nBirdsCreated) return;

	CMatrix m;
	int v4, v58, v56;
	float v55, v52, v59, v60;
	float a3, a4;
	RwV3D outPoint;
	RwV3D point;
	RwRGBA polyColor;
	unsigned char birdAlphaLevel;
	RxObjSpace3dVertex* pVerts; pVerts = (RxObjSpace3dVertex*)(0xC4D958);

	*numIndices = 0;
	*numVerts = 0;

	v4 = 0, v58 = 0, v56 = 0;
	CBird* LastIterationBird = nullptr;

	for(CBird* bird = CBirds::BirdPool; bird != CBirds::PoolEnd; ++bird)
	{
		if(bird->HasBirdHere)
		{
			m.matrixPtr = 0;
			m.haveRwMatrix = 0;

			CMatrix__initIdentityMatrixAtPoint(&m, 0, bird->pos);

			if(bird->field_38 == 1 || bird->field_38 == 2)	// Always true???
			{
				m.matrix.pos.z += (float)(
					sin((long double)((unsigned int)(v4 + *currentTime) % bird->WingSpeed) *
					6.28000020980835 / (long double)bird->WingSpeed) * 0.1000000014901161);
			}

			if(ConvertGlobalXYToScreenXY(&m.matrix.pos, &outPoint, &a3, &a4, 0, 1))
			{
				if(LastIterationBird == nullptr) LastIterationBird = bird;
				else if(LastIterationBird->m_Type != bird->m_Type)
					CBirds::RenderBirdNow(LastIterationBird->m_Type);

				float angle = bird->angle;

				m.matrix.at.z = bird->Size;
				m.matrix.at.x = m.matrix.at.y = 0;
				m.matrix.right.z = m.matrix.top.z = 0;

				m.matrix.right.x = m.matrix.top.y = cos(angle) * bird->Size;
				m.matrix.right.y = m.matrix.top.x = sin(angle) * bird->Size;
				m.matrix.right.y = -m.matrix.right.y;

				if(bird->field_38 == 1 || bird->field_38 == 2)	// Always true???
				{

					// Setup indices
					{
						int* birdIndices = (int*)0x8D52F8;	// Bird indices
						short* rIndices = (short*)((char*)0xC4B958  + (2 * *numIndices));

						for(int i = 0; i < 30; ++i)
						{
							rIndices[i + 0] = *numVerts + birdIndices[i];
							rIndices[i + 30] = *numVerts + birdIndices[i] + BIRD_VERTEX_COUNT;
						}
					}

					// Setup wings
					{
						long double v17 = 6.28000020980835 / (long double)bird->WingSpeed;

						long double v51 = (-sin(((*currentTime + v4 + bird->WingSpeed / 6) % bird->WingSpeed) * v17))
							/ (bird->Size >= 1.0? bird->Size : 1.0);

						long double v20 = (-sin(v17 * (long double)((v4 + *currentTime) % bird->WingSpeed)))
							/ (bird->Size >= 1.0? bird->Size : 1.0);

						v55 = (float)(cos(v51) * 0.5);
						v52 = (float)(sin(v51) * 0.5);
						v59 = (float)(cos(v20) * 0.5);
						v60 = (float)(sin(v20) * 0.5);
					}

					// Get bird alpha level!
					if(!bird->HitByFire){
						RwV3D* CamPos;
						if(camera->__parent.m_pCoords)
							CamPos = &(reinterpret_cast<Placement*>(((char*)camera->__parent.m_pCoords + 0x30))->pos);
						else
							CamPos = &camera->__parent.placement.pos;

						auto v25 = bird->pos.z - CamPos->z;
						auto v26 = bird->pos.y - CamPos->y;
						auto v27 = bird->pos.x - CamPos->x;

						auto v24 = sqrt(v27 * v27 + v26 * v26 + v25 * v25);
						auto v23 = bird->MaxBirdDistance * 0.699999988079071;

						if (v24 > v23)
						{
							auto v28 = 1.0 - (v24 - v23) / (bird->MaxBirdDistance * 0.300000011920929);
							if(v28 < 0.0)
								v28 = 0.0;
							birdAlphaLevel = (unsigned char)(v28 * 255.0);
						}
						else birdAlphaLevel = 255;
					}
					else
					{
						bool foundGround;
						float gZ = FindGroundZFor3DCoord(bird->pos.x, bird->pos.y, bird->pos.z, &foundGround, nullptr);
						if(!foundGround)
						{
							CBirds::DestroyBird(bird);
							continue;
						}

						birdAlphaLevel = (unsigned char)
							(((bird->pos.z - gZ) / (bird->KilledPos.z - gZ)) * 255.0);
					}


					polyColor.alpha = birdAlphaLevel;

					for(int i = 0; i < BIRD_VERTEX_COUNT; ++i)
					{
						bool SetupCustomUV = false;

						switch(i)
						{

							case 8: // New vertex, replaces the top vertex of the bird body.
							case 7: // /|. -- right bottom vertex of the bird body
								SetupCustomUV = true;
								// continue to the next case block

									//  .
							case 0:	// /|  -- replaced with vertex 8!! Now it's just used for the wing
							case 1:	//     -- bottom wing vertex in the middle of the bird body
							
								point.x = 0;
								point.z = 0;
								polyColor.red = bird->PolyColor1;
								polyColor.green = bird->PolyColor2;
								polyColor.blue = bird->PolyColor3;
								break;

							case 6:	// ./| -- left bottom of the bird body
								point.x = (float)1.00000001490116119384765625E-1;
								point.z = 0;
								polyColor.red = bird->PolyColor1;
								polyColor.green = bird->PolyColor2;
								polyColor.blue = bird->PolyColor3;

								SetupCustomUV = true;
								break;

						// Wing vertexes are below

							case 2:
							case 3:
								point.x = v55;
								point.z = v52;
								polyColor.red = bird->PolyColor1;
								polyColor.green = bird->PolyColor2;
								polyColor.blue = bird->PolyColor3;
								break;

							case 4:
							case 5:
								point.x = v59 + v55;
								point.z = v60 + v52;
								polyColor.red = bird->PolyColor4;
								polyColor.green = bird->PolyColor5;
								polyColor.blue = bird->PolyColor6;
								break;
						}

						if(i != 8) point.y = *((float*)0x8D52D8 + i);
						else point.y = *((float*)0x8D52D8 + 0);

						int vidx = (i + *numVerts);
						RxObjSpace3dVertex* LeftVert = &pVerts[0 + vidx];
						RxObjSpace3dVertex* RightVert = &pVerts[BIRD_VERTEX_COUNT + vidx];

						// Setup vertex position
						transformPoint(&outPoint, &m, &point);
						RwIm3DVertexSetPos(LeftVert, outPoint.x, outPoint.y, outPoint.z);
						point.x = -point.x;	// The other side
						transformPoint(&outPoint, &m, &point);
						RwIm3DVertexSetPos(RightVert, outPoint.x, outPoint.y, outPoint.z);

						// Setup vertex RGBA
						RwIm3DVertexSetRGBA(LeftVert, polyColor.red, polyColor.green,
							polyColor.blue, polyColor.alpha);
						RwIm3DVertexSetRGBA(RightVert, BYTE(polyColor.red * 0.8), BYTE(polyColor.green * 0.8),
							BYTE(polyColor.blue * 0.8), polyColor.alpha );
						
						// Setup texture coordinates (UV)
						if(!SetupCustomUV)
						{
							// Setup UV normally, is original vertex
							RwIm3DVertexSetU(LeftVert, *((float*)0x8D52B8 + i) );
							RwIm3DVertexSetV(LeftVert, *((float*)0x8D5298 + i) );
							RwIm3DVertexSetU(RightVert, *((float*)0x8D52B8 + i) );
							RwIm3DVertexSetV(RightVert, *((float*)0x8D5298 + i) );
						}
						else
						{
							// Setup body UV
							float u, v;
							switch(i)
							{
								case 6: u = 0.92578125, v = 0.326171875;	break;
								case 7: u = 0.92578125, v = 0.03125;		break;
								case 8: u = 0.39453125, v = 0.03125;		break;
							}
							RwIm3DVertexSetU(LeftVert, u);
							RwIm3DVertexSetV(LeftVert, v);
							RwIm3DVertexSetU(RightVert, u);
							RwIm3DVertexSetV(RightVert, v);
						}

						if(bird->HitByFire && birdAlphaLevel <= 0.0)
							CBirds::DestroyBird(bird);

					}
					v4 = v56;
					*numIndices += 60;
					*numVerts += BIRD_VERTEX_COUNT*2;
					LastIterationBird = bird;
				}


			}
			CMatrix__destructor(&m);
		}

		v4 += 100;
		v56 = v4;
	}

	if(*numIndices != 0 && LastIterationBird != nullptr)
		CBirds::RenderBirdNow(LastIterationBird->m_Type);
}


void __cdecl CBirds::RenderBirdNow(int index)
{
	if(*numIndices != 0)
	{
		void (*sub_722150)() = (decltype(sub_722150))( 0x722150 );
		RwTexture* texture = BirdTypes[index].m_Texture;

		RwRenderStateSet(rwRENDERSTATESRCBLEND, (void*) 5);
		RwRenderStateSet(rwRENDERSTATEDESTBLEND, (void*) 6);
		RwRenderStateSet(rwRENDERSTATETEXTUREADDRESS, (void*) 3);
		RwRenderStateSet(rwRENDERSTATEFOGENABLE, (void*) 0);
		RwRenderStateSet(rwRENDERSTATEVERTEXALPHAENABLE, (void*) 1);
		if(texture)
			RwRenderStateSet(rwRENDERSTATETEXTURERASTER, (void*) texture->raster);
		else
			RwRenderStateSet(rwRENDERSTATETEXTURERASTER, (void*) nullptr);

		sub_722150();
	}
}


void __cdecl CBirds::DestroyBird(CBird* Bird)
{
	if(Bird->HasBirdHere)
	{
		*nBirdsCreated -= 1;
		Bird->HasBirdHere = false;

		if(Bird->IsDebugBird)
		{
			CBirds::IsInDebug = false;
			Bird->IsDebugBird = false;
		}
	}
}

void __cdecl CBirds::DestroyAllBirds()
{
	for(auto It = BirdPool; It != PoolEnd; ++It)
		DestroyBird(It);
}

CBird* __cdecl CBirds::CreateDebugBird(float x, float y, float z, int type, const char* typeIdentifier, int randSpeed)
{
	DestroyAllBirds();

	if(type == -1 && typeIdentifier)
		type = CBirds::GetBirdIndexByIdentifier(typeIdentifier);
	if(type == -1)
		return nullptr;

	RwV3D Pos, Dest;
	Pos.x = x; Pos.y = y; Pos.z = z;
	Dest.x = 0; Dest.y = 0; Dest.z = 0;
	
	CBirds::Add(Pos, Dest, 1, type, false);
	auto& Bird = CBirds::BirdPool[0];
	if(!Bird.HasBirdHere) return nullptr;

	Bird.IsDebugBird = true;
	CBirds::IsInDebug = true;
	
	Bird.Velocity.x = Bird.Velocity.y = Bird.Velocity.z = 0;
	Bird.field_C = Bird.Velocity;
	CBirds::SetBirdtype(&Bird, type, typeIdentifier, randSpeed);

	Bird.IsMoveableBird = false;
	return &Bird;
}

void __cdecl CBirds::SetBirdtype(CBird* bird, int birdType, const char* typeIdentifier, int randSpeed)
{

	if(bird->HasBirdHere)
	{
		if(birdType == -1 && typeIdentifier)
			birdType = CBirds::GetBirdIndexByIdentifier(typeIdentifier);
		if(birdType == -1)
			return;

		if(randSpeed == -1)
			randSpeed = bird->m_randSpeed;
		else
			bird->m_randSpeed = randSpeed;

		bird->m_Type = birdType;
		auto& Type = BirdTypes[birdType];

		long double vx = (long double)randSpeed * 0.01999999955296516 + Type.Speed;
		CBirds::SetSpeed(bird, Type.Speed, -1);


		if(Type.WingSpeedDiv == -1)
			bird->WingSpeed = 1000000;
		else
		{// TESTME
			bird->WingSpeed = Type.WingSpeedDiv > (randSpeed*6)? Type.WingSpeedDiv - randSpeed*6: Type.WingSpeedDiv;
		}

		bird->MustDoCurves = false;

		bird->Size = randomf(Type.Size.Min, Type.Size.Max);
		if(!bird->IsDebugBird)
			bird->MaxBirdDistance = Type.Distance;
		else
			bird->MaxBirdDistance = 200;

		bird->PolyColor1 = bird->PolyColor2 = bird->PolyColor3 =
		bird->PolyColor4 = bird->PolyColor5 = bird->PolyColor6 = 0xFF;
	}	
}

void __cdecl CBirds::ReloadBird(CBird* bird)
{
	SetBirdtype(bird, bird->m_Type, nullptr, -1);
}

void __cdecl CBirds::SetBirdSize(CBird* bird, float size)
{
	bird->Size = size;
}

void __cdecl CBirds::SetSpeed(CBird* bird, float speed, int randSpeed)
{
	if(randSpeed == -1)
		randSpeed = bird->m_randSpeed;
	else
		bird->m_randSpeed = randSpeed;

	if(speed == -1.0)
		speed = BirdTypes[bird->m_Type].Speed;
	else
		BirdTypes[bird->m_Type].Speed = speed;

	long double vx = (long double)randSpeed * 0.01999999955296516 + speed;
	RwV3D v;
	v.x = sin(bird->angle);
	v.y = cos(bird->angle);
	v.z = bird->Velocity.z;
	CVector__normalize(&v);

	bird->Velocity.x = (float)(v.x * vx);
	bird->Velocity.y = (float)(v.y * vx);
	bird->Velocity.z = (float)(v.z * vx);
	bird->field_C = bird->Velocity;
}

void __cdecl CBirds::SetWingSpeed(CBird* bird, int wingSpeedDiv, int randSpeed)
{
	auto& Type = BirdTypes[bird->m_Type];
	if(randSpeed == -1)
		randSpeed = bird->m_randSpeed;
	else
		bird->m_randSpeed = randSpeed;

	if(Type.WingSpeedDiv == -1)
		bird->WingSpeed = 1000000;
	else
	{
		bird->WingSpeed = Type.WingSpeedDiv > (randSpeed*6)? Type.WingSpeedDiv - randSpeed*6: Type.WingSpeedDiv;
	}
}


int __cdecl CBirds::GetBirdIndexByIdentifier(const char* id)
{
	// Find bird index
	for(auto It = BirdTypes.begin(); It != BirdTypes.end(); ++It)
		if(!_stricmp(It->Identifier, id))
			return It - BirdTypes.begin();
	return -1;
}

int __cdecl CBirds::AddNewBirdtype(BirdType* ty)
{
	static int callCounter = -1;
	++callCounter;

	if(*ty->Identifier == 0) sprintf(ty->Identifier, "NONAME%d", callCounter);
	BirdTypes.push_back(*ty);

	return BirdTypes.size()-1;
}

BirdType* __cdecl CBirds::GetBirdtypePtr(int idx)
{
	return &BirdTypes[idx];
}

int __cdecl CBirds::GetRandomBirdType()
{
	for(int Try = 0; Try < 10; ++Try)
	{
		int i = random(0, BirdTypes.size());
		if(*BirdTypes[i].Identifier != 0) return i;
	}

	return -1;
}


void __cdecl CBirds::KillBird(CBird* bird)
{
	// I ain't good with physics, this is my best! If you're good at physics you can modifie the mod and send to me!

	bird->HitByFire = true;
	bird->WingSpeed = 1000000;

	float SavedSpeed = BirdTypes[bird->m_Type].Speed;
	CBirds::SetSpeed(bird, 5.0, -1);
	BirdTypes[bird->m_Type].Speed = SavedSpeed;
	bird->Velocity.z = -4.0;
	bird->MustDoCurves = true;

	bird->KilledPos = bird->pos;
}






















bool  __cdecl CBirds::RewriteBirdData()
{
	return RewriteBirdsData();
}