/**
	The Birds Update
	ASI Plugin	(Programming)	- by LINK/2012		(http://www.link2012world.blogspot.com/)
	New Birds	(Textures)		- by Junior_Djjr	(http://juniordjjr.blogspot.com.br/)
*/
#pragma once
#include "Injector.h"
#include "RwV3D.h"
#include "NewBirds.h"

struct CBird_Original
{
	RwV3D pos;						// Bird position
	RwV3D field_C;					// Velocity in the current frame
	RwV3D Velocity;					// Velocity (constant if the bird is not doing curves)
	float angle;					// Bird angle

	unsigned int field_28;			// Always 0
	
	unsigned int WingSpeed;			//
	float Size;						// Determins the bird size

	float MaxBirdDistance;			// When the bird is in this distance from the camera, it will disappear.
									// Used also in rendering process to determine the alpha level of the bird.

	char field_38;					// Always 1, if different than 1 or 2 the bird will not be rendered
	
	// Polygons color
	unsigned char PolyColor1;
	unsigned char PolyColor2;
	unsigned char PolyColor3;
	unsigned char PolyColor4;
	unsigned char PolyColor5;
	unsigned char PolyColor6;

	bool HasBirdHere;			// This flags indicates if in this index there's a bird created
	bool MustDoCurves;			// If this flag is true the bird will do curves

	// Unused - padding?
	char field_41;
	char field_42;
	char field_43;
};

struct CBird : public CBird_Original
{
	int m_Type;

	bool IsDebugBird;
	bool IsMoveableBird;
	bool HitByFire;
	bool _b3;

	int m_randSpeed;
	RwV3D KilledPos;
};


static_assert(sizeof(CBird_Original) == 0x44 &&
	sizeof(CBird) == sizeof(CBird_Original) + 24,
	"Incorrect struct size: CBird");


class CBirds
{
	public:
		// Const
		static const int PoolSize = 6;

		// Variables pointers
		static CBird BirdPool[6];
		static CBird* PoolEnd;
		static RwV3D* LastKilledBirdPos;
		static bool* HasKilledAnyBird;
		static int* nBirdsCreated;

		static float* XOffset;
		static float* YOffset;
		static float* ZOffset;

		static void* MethodTable[];

		// Original methods
		static void __cdecl Init();
		static void __cdecl UnInit();
		static int __cdecl Add(RwV3D pos, RwV3D dest, int numBirds, int birdType, bool checkObstacles, ...);
		static void __cdecl Process();
		static void __cdecl Render();
		static void __cdecl CheckIfHitByFire(RwV3D* p1, RwV3D* p2);

		// Additional methods
		static int __cdecl GetRandomBirdType();

		static int __cdecl GetBirdIndexByIdentifier(const char* id);

		static void __cdecl RenderBirdNow(int index);
		static void __cdecl DestroyBird(CBird* Bird);
		static void __cdecl DestroyAllBirds();
		static void __cdecl SetBirdtype(CBird* Bird, int newType, const char* id, int randSpeed);
		static void __cdecl ReloadBird(CBird* bird);
		static int __cdecl AddNewBirdtype(BirdType* ty);
		static BirdType* __cdecl GetBirdtypePtr(int idx);
		static bool  __cdecl RewriteBirdData();

		static void __cdecl SetBirdSize(CBird* bird, float size);
		static void __cdecl SetWingSpeed(CBird* bird, int wingSpeedDiv, int randSpeed);
		static void __cdecl SetSpeed(CBird* bird, float speed, int randSpeed);
		static void __cdecl KillBird(CBird* bird);
		

		static size_t StatusSaver;
		static void __cdecl ReloadBirdTextures()
		{
			ReadBirdsTextures(BirdTypes);
		}
		static void __cdecl SaveDataStatus()
		{
			StatusSaver = BirdTypes.size();
		}
		static void __cdecl RestoreDataStatus()
		{
			if(StatusSaver != BirdTypes.size())
				BirdTypes.resize(StatusSaver);
		}


		
		static bool IsInDebug;
		static CBird* __cdecl CreateDebugBird(float x, float y, float z, int type, const char* id, int randSpeed);
};


inline void CBirds_InitMethods()
{
	//CBirds::BirdPool = (CBird*) 0xC6A8B0;
	CBirds::LastKilledBirdPos = (RwV3D*) 0xC6AA48;
	CBirds::HasKilledAnyBird = (bool*) 0xC6A8A0;
	CBirds::nBirdsCreated = (int*) 0xC6A8A4;

	CBirds::PoolEnd = &CBirds::BirdPool[CBirds::PoolSize];

	CBirds::XOffset = (float*) 0x8D5250;
	CBirds::YOffset = (float*) 0x8D5268;
	CBirds::ZOffset = (float*) 0x8D5280;
}

extern "C"  __declspec(dllexport) void* GetCBirdsMethodTable();





