/**
	by LINK/2012 (http://www.link2012world.blogspot.com/)
	--------- Inspired by the cool CLEO CodeInjector header --------------
*/
#pragma once
#include <Windows.h>
#include <cstdint>
#include "Hooks.h"

//
//
//
union memory_pointer
{
	void*	 p;
	uintptr_t a;

	memory_pointer()						{ p = nullptr; }
	memory_pointer(void* x) : p(x)			{}
	memory_pointer(uint32_t x) : a(x)		{}

	operator void*()		{ return p; }
	operator uintptr_t()	{ return a; }

	memory_pointer& operator=(void* x)		{ return p = x, *this; }
	memory_pointer& operator=(uintptr_t x)	{ return a = x, *this; }
};


//
//
//
extern void WriteMemoryRaw(memory_pointer Address, void* value, size_t size, bool vp);
extern void ReadMemoryRaw(memory_pointer Address, void* ret, size_t size, bool vp);

template<class T> inline
	void WriteMemory(memory_pointer Address, const T& value, bool vp = false)
	{
		DWORD oldProtect;
		if(vp) UnprotectMemory(Address, sizeof(T), oldProtect);
		*(T*)Address.p = value;
		if(vp) VirtualProtect(Address, sizeof(T), oldProtect, &oldProtect);
	}

template<class T> inline
	T& ReadMemory(memory_pointer Address, T& value, bool vp = false)
	{
		DWORD oldProtect;
		if(vp) UnprotectMemory(Address, sizeof(T), oldProtect);
		value = *(T*)Address.p;
		if(vp) VirtualProtect(Address, sizeof(T), oldProtect, &oldProtect);
		return value;
	}

template<class T> inline
	void WriteMemoryNow(memory_pointer Address, T value, bool vp = false)
	{
		WriteMemory(Address, value, vp);
	}

template<class T> inline
	T ReadMemoryNow(memory_pointer Address, bool vp = false)
	{
		T value;
		return ReadMemory(Address, value, vp), value;
	}


inline bool ProtectMemory(memory_pointer Address, size_t size, DWORD Protect)
{
	return VirtualProtect(Address, size, Protect, &Protect) != 0;
}

inline bool UnprotectMemory(memory_pointer Address, size_t size, DWORD& Out_OldProtect)
{
	return VirtualProtect(Address, size, PAGE_EXECUTE_READWRITE, &Out_OldProtect) != 0;
}



//
//
//
enum game_t { III, VC, SA };
class GameVersion
{
	public:
		enum eVersion
		{
			UNKNOWN = -1,
			SA_10US, SA_10EU, SA_101EU
		};

	private:
		game_t m_Game;
		eVersion m_Version;
	public:
		GameVersion(game_t game = SA) : m_Game(game), m_Version(GameVersion::UNKNOWN)
		{}

		GameVersion(const GameVersion& v)
		{ *this = v; }

		bool IsUnknown()	{ return m_Version == this->UNKNOWN; }
		game_t Game()		{ return m_Game; }
		eVersion Version()	{ return m_Version; }

		operator eVersion() { return Version(); }

		GameVersion& operator=(const GameVersion& v)
		{
			this->m_Game = v.m_Game;
			this->m_Version = v.m_Version;
			return *this;
		}

		GameVersion& operator=(eVersion v) { return m_Version = v, *this; }
		bool operator==(eVersion v) { return m_Version == v; }
		bool operator!=(eVersion v) { return !operator==(v); }

		GameVersion& Get();
		const char* GetString();
};


//
//
//
inline void MakeJMP(memory_pointer at, memory_pointer dest)
{
	WriteMemoryNow<uint8_t>(at, 0xE9, true);
	WriteMemoryNow<uint32_t>(at+1, dest - (at+5), true);
}

inline void MakeCALL(memory_pointer at, memory_pointer dest)
{
	WriteMemoryNow<uint8_t>(at, 0xE8, true);
	WriteMemoryNow<uint32_t>(at+1, dest - (at+5), true);
}

inline void MakeJA(memory_pointer at, memory_pointer dest)
{
	WriteMemoryNow<uint16_t>(at, 0x87F0, true);
	WriteMemoryNow<uint32_t>(at+2, dest - (at+5), true);
}

