/**
	The Birds Update
	ASI Plugin	(Programming)	- by LINK/2012		(http://www.link2012world.blogspot.com/)
	New Birds	(Textures)		- by Junior_Djjr	(http://juniordjjr.blogspot.com.br/)
*/
#define _CRT_SECURE_NO_WARNINGS

#include "NewBirds.h"
#include "Injector.h"
#include "Addresses.h"
#include "Hooks.h"
#include "CBirds.h"

BirdTypeVector BirdTypes;

FILE* logfile;

char* GetDataLine(char* Buffer, int maxCount, FILE* f)
{
	while(fgets(Buffer, maxCount, f))
	{
		if(Buffer[0] == '\n' || Buffer[0] == '#') continue;

		for(char* p = Buffer; *p; ++p)
		{
			if(*p == ',') *p = ' ';
			else if(*p == '\n' || *p == '#' || *p == ';')
			{
				*p = 0;
				break;
			}
		}

		if(Buffer[0] != 0) return Buffer;
	}

	return nullptr;
}


BirdTypeVector&
	ReadBirdsData(const char* dataFilename, BirdTypeVector& result)
{
	BirdType br;
	std::vector<BirdType> NoIDBirds;

	// Buffers
	char Identifier[16+32];
	char SpawningTown[8+32];
	char Preferences[32+32];
	char line[512];

	//
	FILE* f;
	int HighestID = -1;

	if(! (f = fopen(dataFilename, "r")) ) return result;

	while(GetDataLine(line, sizeof(line), f))
	{
		if(sscanf(line, "%s %d %s %d %d %d %d %f %f %f %f %f %f %d %s",
			Identifier, &br.ForceType, SpawningTown, &br.SpawnTime.Min, &br.SpawnTime.Max,
			&br.BirdsInFlock.Min, &br.BirdsInFlock.Max, &br.Height.Min, &br.Height.Max,
			&br.Size.Min, &br.Size.Max, &br.Distance, &br.Speed, &br.WingSpeedDiv, Preferences)
		< 15) continue;
		
		Identifier[15] = 0;
		SpawningTown[5] = 0;

		strcpy(br.Identifier, Identifier);
		
		if(br.BirdsInFlock.Min < 1) br.BirdsInFlock.Min = 1;
		if(br.BirdsInFlock.Max > CBirds::PoolSize+1) br.BirdsInFlock.Max = CBirds::PoolSize;
		
		br.Flags = 0x0;
		for(int i = 0; i < 5; ++i)
			br.SpawningTown[i] = SpawningTown[i] != '0';
		for(int i = 0; Preferences[i] != 0; ++i)
			br.Flags |= (Preferences[i] != '0') << i;

		br.m_Texture = nullptr;


		if(!_stricmp(Identifier, "DOVE"))
			br.ForceType = 0;
		if(!_stricmp(Identifier, "HAWK"))
			br.ForceType = 1;
		if(!_stricmp(Identifier, "SEAGULL"))
			br.ForceType = 2;

		if(br.ForceType <= -1) NoIDBirds.push_back(br);
		else
		{
			if(br.ForceType > HighestID) HighestID = br.ForceType;
			if(result.size() <= (size_t)br.ForceType) result.resize(br.ForceType+64);
			result[br.ForceType] = br;
		}
	}

	fclose(f);

	result.resize(HighestID+1);
	if(NoIDBirds.size() != 0)
		result.insert(result.end(), NoIDBirds.begin(), NoIDBirds.end());
	CBirds::StatusSaver = result.size(); 

	/*fprintf(logfile, "Birds:\n");
	for(unsigned int i = 0; i < result.size(); ++i)
		fprintf(logfile, "%d = %s\n", i, result[i].Identifier);*/


	return result;
}

bool RewriteBirdsData()
{
	char* buffer, *p;
	FILE* f;
	long size;
	chdirProgramDir("");

	// Take a copy of the comments into buffer
	{
		f = fopen("data\\birds.dat", "rb");
		if(!f) return false;

		fseek(f, 0, SEEK_END);
		size = ftell(f);
		rewind(f);

		buffer = new char[size+6];
		buffer[size+1] = buffer[size+2] = buffer[size+3] = 0;
		fread(buffer, sizeof(char), size, f);
		fclose(f);

		p = buffer;
		while(*p == '#')
		{
			char* pp = strchr(p, '\n');
			if(pp)
			{
				if(*(pp - 1) == '\r') *(pp - 1) = ' ';
			}
			
			p = pp+1;
		}
		*p = 0;
	}

	// Write the new data file
	{
		f = fopen("data\\birds.dat", "w");
		if(!f)
		{
			delete[] buffer;
			return false;
		}

		if(p - buffer != 0)
		{
			fwrite(buffer, sizeof(char), p - buffer, f);
			fputc('\n', f);
		}

		char SpawningTown[32], Preferences[32];
		SpawningTown[5] = 0;

		for(auto It = BirdTypes.begin(); It != BirdTypes.end(); ++It)
		{
			auto& br = *It;

			for(int i = 0; i < 5; ++i)
				SpawningTown[i] = br.SpawningTown[i]? '1' : '0';
			
			unsigned int ff = 1, c = 0;
			for(; ff < BIRD_MAX_FLAG; ff <<= 1, ++c)
				Preferences[c] = GetFlag(br.Flags, ff)? '1' : '0';
			Preferences[c] = 0;
	
			fprintf(f, "%s\t%d\t%s\t%d\t%d\t%d\t%d\t%f\t%f\t%f\t%f\t%f\t%f\t%d\t%s\n",
					br.Identifier, br.ForceType, SpawningTown, br.SpawnTime.Min, br.SpawnTime.Max,
					br.BirdsInFlock.Min, br.BirdsInFlock.Max, br.Height.Min, br.Height.Max,
					br.Size.Min, br.Size.Max, br.Distance, br.Speed, br.WingSpeedDiv, Preferences
				);
		}

		delete[] buffer;
		fclose(f);
	}
	return true;
}


BirdTypeVector&
	ReadBirdsTextures(BirdTypeVector& br)
{
	WIN32_FIND_DATAA FindData;
	HANDLE FindHandle;
	int txdIndex;

	DestroyBirdsDictionary(br);

	txdIndex = txdIndexByName("birds");
	if(txdIndex == -1) txdIndex = txdAdd("birds");
	txdAddRef(txdIndex);

	chdirProgramDir("models\\birds");
	FindHandle = FindFirstFileA("*.txd", &FindData);
	if(FindHandle == INVALID_HANDLE_VALUE) return br;

	//fprintf(logfile, "Scanning TXD's\n");
	do
	{
		txdLoad(txdIndex, FindData.cFileName);
	}
	while(FindNextFileA(FindHandle, &FindData));

	for(auto It = br.begin(); It != br.end(); ++It)
	{
		txdSetCurrent(txdIndex);
		It->m_Texture = RwReadTexture(It->Identifier, nullptr);
	}

	chdirProgramDir("");
	FindClose(FindHandle);
	return br;
}


BirdTypeVector&
	GetBirdsData(const char* dataFilename, BirdTypeVector& result)
{
	//logfile = fopen("result.txt", "w");
	ReadBirdsTextures( ReadBirdsData(dataFilename, result) );
	//fclose(logfile);
	return result;
}


void DestroyBirdsDictionary(BirdTypeVector& br)
{
	for(auto It = br.begin(); It != br.end(); ++It)
	{
		if(It->m_Texture)
		{
			RwTextureDestroy(It->m_Texture);
			It->m_Texture = nullptr;
		}
	}

	int txdIndex = txdIndexByName("birds");
	if( txdIndex != -1 ) txdRemove(txdIndex);
}
