#pragma once
#include "Injector.h"
#include "RwV3D.h"

struct CBird
{
	RwV3D pos;						// Bird position
	RwV3D field_C;					// Velocity in the current frame
	RwV3D Velocity;					// Velocity (constant if the bird is not doing curves)
	float angle;					// Bird angle

	unsigned int field_28;			// Always 0
	
	int WingSpeed;					//
	float Size;						// Determins the bird size

	float MaxBirdDistance;			// When the bird is in this distance from the camera, it will disappear.
									// Used also in rendering process to determine the alpha level of the bird.

	char field_38;					// Always 1, if different than 1 or 2 the bird will not be rendered
	
	// Polygons color
	unsigned char PolyColor1;
	unsigned char PolyColor2;
	unsigned char PolyColor3;
	unsigned char PolyColor4;
	unsigned char PolyColor5;
	unsigned char PolyColor6;

	bool HasBirdHere;			// This flags indicates if in this index there's a bird created
	bool MustDoCurves;			// If this flag is true the bird will do curves

	// Unused - padding?
	char field_41;
	char field_42;
	char field_43;
};

ASSERT_SIZE(CBird, 0x44);


class CBirds
{
	public:
		// Const
		static const int PoolSize = 6;

		// Variables pointers
		static CBird* BirdPool;
		static CBird* PoolEnd;
		static RwV3D* LastKilledBirdPos;
		static bool* HasKilledAnyBird;
		static int* nBirdsCreated;

		static float* XOffset;
		static float* YOffset;
		static float* ZOffset;

		static void __cdecl Init();
		static void __cdecl Shutdown();
		
		static int __cdecl Add(RwV3D pos, RwV3D dest, int numBirds, int birdType, bool checkObstacles);

		static void __cdecl Process();
		static void __cdecl Render();
		
		static void __cdecl CheckIfHitByFire(RwV3D* p1, RwV3D* p2);
};


inline void CBirds_InitMethods()
{
	CBirds::BirdPool = (CBird*) 0xC6A8B0;
	CBirds::LastKilledBirdPos = (RwV3D*) 0xC6AA48;
	CBirds::HasKilledAnyBird = (bool*) 0xC6A8A0;
	CBirds::nBirdsCreated = (int*) 0xC6A8A4;

	CBirds::PoolEnd = &CBirds::BirdPool[CBirds::PoolSize];

	CBirds::XOffset = (float*) 0x8D5250;
	CBirds::YOffset = (float*) 0x8D5268;
	CBirds::ZOffset = (float*) 0x8D5280;
}





